## Method of moments estimator for Fst based on
## data with alleles by rows and populations by column
"MoMFst" <- function(m)
{
  n <- colSums(m)
  n.alleles <- nrow(m)
  n.pops <- ncol(m)
  p <- m/n                         # relative allele frequencies
  p.bar <- rowSums(m)/sum(m);      # overall allele frequencies
  pm <- (p-matrix(p.bar,ncol=n.pops,nrow=n.alleles))^2
  s <- rowSums(pm)/p.bar*(1-p.bar)
  mean(s)
}


## Weir and Cockerham's estimator for Fst based on
## data with alleles by rows and populations by column
## A direct copy from my C code
"WCFsta" <- function(m,transpose=FALSE)
  {
    if (transpose) m <- t(m)

    npops <- ncol(m)
    n <- colSums(m)
    nb <- mean(n)
    nc <- (sum(n)-(sum(n^2)/sum(n)))/(npops-1)
 #   print(nc)

    f <- apply(m,2,function(x)  x/sum(x))
    r <- apply(f,1,function(x) {
      m <- mean(x)
      v <- var(x)
      a <- (nb/nc)*(v
                    -(1/(nb-1))
                    *( m*(1-m)-((npops-1)/npops)*v
                      )
                    )
      d <- nb/(nb-1)*(m*(1-m)-((npops-1)/npops)*v)
      c(a,d)
    })
#    print(r)
   d <- rowSums(r)
   d[1]/(d[1]+d[2])
  }
  ## A version of the Fst estimator taken from Weir2005 et al 
"WCFst" <- function(m,transpose=FALSE)
  {
    if (transpose) m <- t(m)

    npops <- nrow(m)
    p <- apply(m,2,function(x)  x/sum(x))
    r <- apply(f,1,function(x) {
      m <- mean(x)
      v <- var(x)
      a <- (nb/nc)*(v
                    -(1/(nb-1))
                    *( m*(1-m)-((npops-1)/npops)*v
                      )
                    )
      d <- nb/(nb-1)*(m*(1-m)-((npops-1)/npops)*v)
      c(a,d)
    })
#    print(r)
   d <- rowSums(r)
   d[1]/(d[1]+d[2])
  }
  
