## helper function for Weir Hill Fst Calculations
WH <- function(x)
  {
    if (class(x)!="table")
      stop("WH expects a table as input")
    n.pops <- ncol(x)
    n <- colSums(x)
    tot.n <- sum(n)
    n.ic <- n-n^2/sum(n)
    n.c <- sum(n.ic)/(n.pops-1)
   
    n.alleles <- nrow(x)
    p.bar <- rowSums(x)/sum(x)      # overall allele frequencies
    p.bar.mat <- matrix(p.bar,ncol=n.pops,nrow=n.alleles)
    p <- x/n                        # relative allele frequencies in each population
    pi.u <- p*(1-p)                 # eqn from p727
    n.mat <- matrix(n,ncol=n.pops,nrow=n.alleles,byrow=T)
 
    MSP <- rowSums(n.mat*(p-p.bar.mat)^2)/(n.pops-1)
    MSG <- rowSums(n.mat*pi.u)/sum(n-1)

    a <- sum(MSP-MSG)
    b <- sum(MSP+(n.c-1)*MSG)
    c(a,b)
  }

"WeirHill" <- function(m)
{
  ## assume that we have a matrix of allele frequencies at a single
  ## locus, with each row representing an allele and each column a population
  if (class(m)[1]=="onelocus")
    y <- WH(table(m))
  else y <- WH(m)
  
  ## equation 5 from Weir & Hill
  y[1]/y[2]
}

"WeirHill.indep" <-  function(m)
  {
    ## now for an independent estimate for each population
    ## this is only valid (I think) if there are more than 2 alleles
    if (class(m)=="onelocus") m <- table(m)
     n <- colSums(m)
     n.pops <- ncol(m)
     n.alleles <- nrow(m)
     p.bar <- rowSums(m)/sum(m)      # overall allele frequencies
     p.bar.mat <- matrix(p.bar,ncol=n.pops,nrow=n.alleles)
     n.mat <- matrix(n,ncol=n.pops,nrow=n.alleles,byrow=T)
     p <- m/n.mat  # relative allele frequencies in each population
     pi.u <- p*(1-p)                 # eqn from p727
     n.c <- n-n^2/sum(n)
     n.c.mat <- matrix(n.c,ncol=n.pops,nrow=n.alleles,byrow=T)
     ## now to calculate equation 6 for Weir and Hill
     
     a <- sum(n.c)*colSums(pi.u)
     b <- n.mat*(p-p.bar.mat)^2+n.c.mat*pi.u

     return (1.-a/sum(b))
  }

WeirHill.ml <- function(m)
  {
    loctab <- function(b) {
      table(b,m$location)
    }
    tabs <- lapply(m$x,loctab)
    r <- rowSums(sapply(tabs,WH))
    r[1]/r[2] 
  }
