
rLGESF <- function(theta,F,n)
### Function to generate a sample of size n (which may be a vector)
### from the combined local subpopulation model (of Balding and Nichols)
### with the Ewens Sampling Formula
  {
    ## utility function
    nextEwens <- function(theta,x) {
      n <- sum(x);
      k <- length(x)+1;
      p <- c(x,theta) # does not need to be normalised/(n+theta);
      a <- sample(k,1,prob=p);
      a;
    }
    ## start with some checks that the input is valid
    if (max(F)>=1.0)
      stop("F must be less than 1")
    if (min(F)<0)
      stop("minimum value of F is 0")
    if (theta<=0.0) stop("theta must be strictly positive")

    ## extract information from input
    n.pops <- length(n)
    tot.n <- sum(n)
    
    if (length(F)==1)
      F <- rep(F,n.pops)       # individual F's for each population
    else if (length(F)!=n.pops)
      stop("F must have length 1 or the same as the sample size")

    ## get a random ordering of the sample
    
    order <- sample(rep(1:n.pops,n))
    ss <- numeric(n.pops)
    max.alleles <- 100
       
    ## now set up the variables after the first sample
    xglobal <- c(1,rep(0,max.alleles-1))            # the global list of alleles
    k <- 1                   # the number of alleles
    ss[order[1]] <- 1        # current sample size from each population
    xlocal <- matrix(0,ncol=max.alleles,nrow=n.pops)
    xlocal[order[1]][1] <- 1
    
    ## loop through the rest of the samples
    for (samp in 2:tot.n) {
     pop <- order[samp]
      if (ss[pop]>0) { ## have we already seen a sample from this population
        plocal <- (F[pop]*xlocal[pop,]/(1+(ss[pop]-1)*F[pop]))[1:k]
        if (runif(1)<sum(plocal)) {
          a <- sample(k,1,prob=plocal)
        } else {
          a <- nextEwens(theta,xglobal[1:k])
          if (a==k+1) k <- k+1
          xglobal[a] <- xglobal[a]+1
        }
      } else {
        a <- nextEwens(theta,xglobal[1:k])
        if (a==k+1) k <- k+1
        xglobal[a] <- xglobal[a]+1
      }
     xlocal[pop,a] <-  xlocal[pop,a]+1
     ss[pop] <- ss[pop]+1
    }

    return(as.table(t(xlocal[,1:k])))
  }
      
    
