#include <iostream>
#include <string>
#include <vector>
#include <sstream>                // for ostringstream
#include "onelocusF.H"
#include "onelocusMCMC.H"
#include "gsl_rand.H"
#include "gsl_distributions.H"




extern "C" {
  void onelocusFimpll(int *x, int *loc, int *n, double *F, int *nF , double *drive
		      , int *nperms, double *theta, double *ll) {

    onelocusF a(x,loc,*n);
    rng r;

    TNT::Array2D<double> results(*nF,*nperms);
    TNT::Array2D<double> Fval(*nF,a.npops);

    for (int i=0;i< *nF;i++) {
      for (int j=0;j<a.npops;j++) Fval[i][j]=F[i];
    }
    TNT::Array1D<double> driveF(a.npops,*drive);

    if (F[0]==0.0) {
      for (int i=0;i<  *nperms;i++) {
	a.perm(r);
	results[0][i]=a.likelihoodF0(*theta);
	a.resampleg_b(driveF,*theta,r);
	double lq=a.likelihood(*theta);
	for (int j=1;j<*nF;j++) {
	  double lq2=a.lpg(Fval[j]);
	  double lp2=a.lpsampleg_b(Fval[j],*theta);
	  // double lp2=a.lpsampleg(Fval[j]);
	  results[j][i] = lq2-lp2+lq;
	}
      }
    } else {
      for (int i=0;i<  *nperms;i++) {
	a.perm(r);
	a.resampleg_b(driveF,*theta,r);
	// a.resampleg(driveF,r);
	double lq=a.likelihood(*theta);
	for (int j=0;j<*nF;j++) {
	  double lq2=a.lpg(Fval[j]);
	  double lp2=a.lpsampleg_b(Fval[j],*theta);
	  //	  std::cerr << lq << " " << lq2 << " " << lp2 << std::endl;
	  // double lp2=a.lpsampleg(Fval[j]);
	  results[j][i] = lq2-lp2+lq;
	}
      }
    }
    
    double logn=log(double(*nperms));
    for (int i=0;i<*nF;i++) {
      double me=*(std::max_element(results[i],results[i]+*nperms));
      ll[i]=0.0;
      for (int j=0;j< *nperms;j++) ll[i] += exp(results[i][j]-me); 
      ll[i]=log(ll[i])-logn+me;
    }
  }
  // now with an F for each population
  void onelocusFimpllb(int *x, int *loc, int *n, double *F, double *drive
		      , int *nperms, double *theta, double *ll) {

    onelocusF a(x,loc,*n);
    rng r;

    TNT::Array1D<double> results(*nperms);
  
    TNT::Array1D<double> driveF(a.npops,*drive);
    for (int i=0;i<a.npops;i++) 
      if (F[i]<=0.0) driveF[i]=0.0;

    for (int i=0;i<  *nperms;i++) {
      a.perm(r);
      a.resampleg_b(driveF,*theta,r);
      // a.resampleg(driveF,r);
      double lq=a.likelihood(*theta);
      double lq2=a.lpg(F);
      double lp2=a.lpsampleg_b(F,*theta);
      results[i] = lq2-lp2+lq;
    }
    
    double logn=log(double(*nperms));
    double me=*(std::max_element(&results[0],&results[0]+*nperms));
    *ll=0.0;
    for (int j=0;j< *nperms;j++) *ll += exp(results[j]-me); 
    *ll=log(*ll)-logn+me;
    
  }

  void onelocusFMCMC(int *x, int *loc, int *n, int *reps, int *thinning
		     , char **thetaprior, char **Fprior
		     , double *F, double *theta) {
    
    // improper prior for theta, uniform prior for F on [0,1], 
    // all populations have the same F
    
    ctsdistribution *tprior=fromstring(std::string(*thetaprior));
    ctsdistribution *fprior=fromstring(std::string(*Fprior));
    double tuneF=0.1;
    double tunet=1.0;

    onelocusMCMC mod(x,loc,*n,tprior,fprior);

    TNT::Array1D<double> success(5,0.0);

    for (int i=0;i< *reps;i++) {
      for (int j=0;j< *thinning;j++) {
	success[0]+= mod.updateF(tuneF);
	success[1]+=mod.updateg3();
	success[2]+=mod.updatetheta(tunet);
	success[3]+=mod.updateg();
      }
      mod.check();
      F[i]=mod.F[0];
      theta[i]=mod.theta;
    }
    
    std::cerr << "success rates:  F - " << success[0]/(double( *reps)*(double) *thinning) 
    	      << "   global1 - " << success[1]/(double( *reps)*(double) *thinning) 
          << "   theta - " << success[2]/(double( *reps)*(double) *thinning) 
	      << "   global2 - " << success[3]/(double( *reps)*(double) *thinning) 
         << std::endl;  
   
  }
}
